package jst.parsetree;
import parser.parsetree.*;
import parser.parsetree.visitor.*;

/** Gof Visitor design pattern variant.
 *  This differs from the standard pattern: The visitor is responsible for navigation.  This is a
 *  known variant, suitable for cases where navigation is complex.
 */
public abstract class ParseTreeVisitor extends SyntaxTreeVisitor {

    // Remove cycle detection for efficiency.
    public void visitBranch(SyntaxTreeBranch branch) {

        visitNode(branch);

        SyntaxTree[] kids = branch.getKids();

        if (branch.getProductionType() == null) {
            SyntaxTreeBranch firstKid = kids[0].asBranch();
            if (firstKid.getProductionType().getLhs().getID() != Symbol.NT_CLASS_OR_INTERFACE_TYPE)
                System.out.println("ambig " + firstKid.getProductionType());

            kids[0].accept(this);
            return;
        }

        for (int i = 0; i < kids.length; i++)
            if (kids[i] != null)
                kids[i].accept(this);
    }

    protected int getType(SyntaxTree node) {
        return node.getSymbolType().getID();
    }

    protected SyntaxTreeLeaf getLeaf(SyntaxTreeBranch branch, int where) {
        SyntaxTree[] kids = branch.getKids();

        if (where >= kids.length)
            return null;

        SyntaxTree kid = kids[where];
        if (kid != null)
            return kid.asLeaf();

        return null;
    }

    protected SyntaxTreeLeaf getLeaf(SyntaxTreeBranch branch, int type, int from, int to) {
        SyntaxTree[] kids = branch.getKids();
        SyntaxTree kid = getNode(kids, type, from, to);
        if (kid != null)
            return kid.asLeaf();

        return null;
    }

    protected SyntaxTreeBranch getBranch(SyntaxTreeBranch branch, int type, int from) {
        return getBranch(branch, type, from, from);
    }

    protected SyntaxTreeBranch getBranch(SyntaxTreeBranch branch, int type, int from, int to) {
        SyntaxTree[] kids = branch.getKids();
        SyntaxTree kid = getNode(kids, type, from, to);
        if (kid != null)
            return kid.asBranch();

        return null;
    }

    protected SyntaxTree getNode(SyntaxTree[] kids, int type, int from, int to) {
        for (int i = from; i <= to && i < kids.length; i++) {
            SyntaxTree kid = kids[i];
            if (getType(kid) == type)
                return kid;
        }

        return null;
    }
}
