
package jst.symtab;

import java.io.*;
import java.util.*;

import jst.parsetree.*;
import java.lang.reflect.Modifier;
import jst.symtab.visitor.ModelVisitor;

public class FieldDecl extends VariableDecl {

	protected int modifier;

    /**
     * @supplierCardinality 0..1
     * @supplierRole initializer 
     * @link aggregation
     */
    protected FieldInitializer initializer;

	public FieldDecl(UserType ownerUserType, String theSimpleName, JstParseTreeBranch theSource) {
    	super(ownerUserType, theSimpleName, theSource);

       	ownerUserType.addField(this);
    }

	public FieldDecl(ArrayType ownerArrayType, String theSimpleName, JstParseTreeBranch theSource) {
    	super(ownerArrayType, theSimpleName, theSource);

       	// ownerArrayType.addField(this);  It already knows.
    }

	public FieldDecl(UserType newOwnerUserType, FieldDecl that, Map<TypeParameter, ActualType> swaps) {
		super(newOwnerUserType, that, swaps);

       	newOwnerUserType.addField(this);

		modifier = that.modifier;

		if (that.initializer != null)
			that.initializer.copy(this, swaps);
	}

	public void copy(UserType newOwnerUserType, Map<TypeParameter, ActualType> swaps) {
		new FieldDecl(newOwnerUserType, this, swaps);
	}

 	public String getCanonicalName() {
       return "FLD_" + getOwnerScope().getCanonicalName().substring(4) + "." + simpleName;
    }

	public int getModifier() {
    	return modifier;
    }

	public void setModifier(int theModifier) {
    	modifier = theModifier;
    }

    public FieldInitializer getInitializer() {
        return initializer;
    }

    public void setInitializer(FieldInitializer theInitializer) {
        initializer = theInitializer;
    }

    public boolean accessible(ReferenceType looker) {
    	if(looker == null) {
    		return true;
    	}
        if (Modifier.isPublic(modifier))
        	return true;

		if (Modifier.isProtected(modifier))
			return getOwnerPackage() == looker.getOwnerPackage() || looker.isSubtypeOf(getOwnerUserType());

		if (Modifier.isPrivate(modifier))
        	return looker.isContainedBy(getOwnerUserType());

		// Package access
		return getOwnerPackage() == looker.getOwnerPackage();
    }

    public void accept(ModelVisitor visitor) {
        visitor.visitFieldDecl(this);
    }
}
