
package jst.symtab;
import java.io.*;
import jst.parsetree.*;
import java.util.*;
import java.lang.reflect.Modifier;

import jst.symtab.visitor.ModelVisitor;

public class MethodDecl extends OperationDecl {

	public MethodDecl(UserType ownerUserType, String theSimpleName, JstParseTreeBranch theSource) {
    	super(ownerUserType, theSimpleName, theSource);

    	ownerUserType.addMethod(this);
    }

	public MethodDecl(ArrayType ownerArrayType, String theSimpleName, JstParseTreeBranch theSource) {
    	super(ownerArrayType, theSimpleName, theSource);

        //ownerArrayType.addMethod(this); It already knows.
    }
	
	public MethodDecl(UserType newOwnerUserType, MethodDecl that, Map<TypeParameter, ActualType> swaps) {
		super(newOwnerUserType, that, swaps);

    	newOwnerUserType.addMethod(this);
	}
	
	public MethodDecl(ArrayType newOwnerArrayType, MethodDecl that, Map<TypeParameter, ActualType> swaps) {
		super(newOwnerArrayType, that, swaps);
	}

	public void copy(UserType newOwnerUserType, Map<TypeParameter, ActualType> swaps) {
		new MethodDecl(newOwnerUserType, this, swaps);
	}

	public void copy(ArrayType newOwnerArrayType, Map<TypeParameter, ActualType> swaps) {
	   new MethodDecl(newOwnerArrayType, this, swaps);
	}

	public ConstructorDecl asConstructor() {
		return null;
	}

	public MethodDecl asMethod() {
		return this;
	}

 	public String getCanonicalName() {
       return "MTH_" + getOwnerScope().getCanonicalName().substring(4) + "." + getSignature();
    }

	public String getSignature() {
    	String signature = getSimpleName() + "(";

		Iterator paramIter = parameters.iterator();
		while (paramIter.hasNext()) {
			ParameterDecl param = (ParameterDecl) paramIter.next();
			TypeDecl paramType = param.getType();
			signature += (paramType == null? "UNKNOWN" : paramType.getCanonicalName());
			if (paramIter.hasNext())
				signature += ",";
		}

		signature += ")";

		return signature;
    }

	public boolean hides(OperationDecl op) {
    	if (parameters.size() != op.getParameters().size()) // Performance optimisation
        	return false;

		if (!getSignature().equals(op.getSignature()))
        	return false;

        // Tricky.  It is not enough to check that my class is derived from the
        // class of the given op.  If a subclass of mine inherits me and inherits
        // op from an interface, we must choose me.
		if (!Modifier.isAbstract(modifier) && Modifier.isAbstract(op.getModifier()))
			return true; // this is concrete, op is abstract.

        return getOwnerUserType().isSubtypeOf(op.getOwnerUserType());
    }
	
    /** New code **************************************************************/
    private MethodDecl overrides;
    private Set overriddenBy = new HashSet();
    private Set allOverriddenBy = new HashSet();
    
    public void setOverrides(MethodDecl md) {
        overrides = md;
        if (md != null) {
            md.addAllOverriddenBy(getAllOverriddenBy());
        }
    }
    
    public MethodDecl getOverrides() {
        return overrides;
    }
    
    public void addOverriddenBy(MethodDecl md) {
        overriddenBy.add(md);
        addAllOverriddenBy(md);
        if (overrides != null) {
            overrides.addAllOverriddenBy(md);
        }
    }
    
    public Set getOverriddenBy() {
        return overriddenBy;
    }
    
    public void addAllOverriddenBy(MethodDecl md) {
        allOverriddenBy.add(md);
    }
    
    public void addAllOverriddenBy(Set mds) {
        allOverriddenBy.addAll(mds);
    }
    
    public Set getAllOverriddenBy() {
        return allOverriddenBy;
    }
    
    public MethodDecl findOverrides() {
        UserType owner = getOwnerUserType();
        
        // lookup this method in this user type
        List paramTypes = new ArrayList();
        Iterator iter = parameters.iterator();
        while (iter.hasNext()) {
            paramTypes.add(((ParameterDecl)iter.next()).getType());
        }
        MethodDecl method = owner.getOverridden(getSimpleName(), paramTypes, owner);

        return method;
    }
    
    /** New code ends *********************************************************/

    public void accept(ModelVisitor visitor) {
        visitor.visitMethodDecl(this);
    }
}
