
package jst.symtab;
import java.util.*;
import jst.parsetree.*;
import jst.symtab.link.*;

import java.lang.reflect.Modifier;

public abstract class OperationDecl extends TypedDecl implements Scope {

	protected int modifier;

    /**
     *@link aggregation
     *      @associates <{jst.symtab.ParameterDecl}>
     * @supplierCardinality 0..*
     * @supplierRole parameters*/
	protected List<ParameterDecl> parameters; // ParameterDecls.

    /**
     * @link aggregation
     * @clientCardinality 0..1
     * @supplierCardinality 0..1 
     * @supplierRole body
     * */
    protected Block body;

    /**
     *@link association
     *      @associates <{ClassType}>
     * @supplierRole throwClasses
     * @supplierCardinality 0..*
     */
    protected Set<TypeLookup> throwClassRefs;

	protected OperationDecl(Decl theOwner, String theSimpleName, JstParseTreeBranch theSource) {
    	super(theOwner, theSimpleName, theSource);
    }

	protected OperationDecl(Decl newOwner, OperationDecl that, Map<TypeParameter, ActualType> swaps) {
		super(newOwner, that, swaps);

		modifier = that.modifier;

		for (ParameterDecl thatParam : that.parameters)
			thatParam.copy(this, swaps);

		that.body.copy(this, swaps);

		for (TypeLookup thatThrow : that.throwClassRefs)
			thatThrow.copy(this);
	}

	protected void init() {
        parameters = new ArrayList<ParameterDecl>();
        body = null;
        throwClassRefs = new HashSet<TypeLookup>();
	}

	public Scope asScope() {
		return this;
	}

	public MethodDecl asMethod() {
		return null;
	}
	
	public ConstructorDecl asConstructor() {
		return null;
	}
	
	// Generate artificial name for anonymous decl in this scope.
    // Operations contain anonymous parameters when they are loaded by reflection.
	public String generateName() {
    	return "@PARAM_" + parameters.size();
    }

	/*
	public UserType getOwnerType() {
    	return (UserType) ownerScope;
	}
	*/

	public void addThrowRef(TypeLookup exceptionType) {
		throwClassRefs.add(exceptionType);
	}
	
    public void addParameter(ParameterDecl param) {
		parameters.add(param);
    }

	public List getParameters() {
    	return parameters;
    }
	
	public boolean addType(UserType type) {
    	return false; // Cannot add inner type directly to an operation.
    }

	public void setBody(Block theBody) {
    	if (body != null)
        	throw new RuntimeException("body already set");

		body = theBody;
    }

	public Block getBody() {
    	return body;
    }

	public int getModifier() {
    	return modifier;
    }

	public void setModifier(int theModifier) {
    	modifier = theModifier;
    }

    public boolean accessible(ReferenceType looker) {
        if (Modifier.isPublic(modifier))
        	return true;

		if (Modifier.isProtected(modifier))
			return getOwnerPackage() == looker.getOwnerPackage() || looker.isSubtypeOf(getOwnerUserType());

		if (Modifier.isPrivate(modifier))
        	return looker.isContainedBy(getOwnerUserType());

		// Package access
		return getOwnerPackage() == looker.getOwnerPackage();
    }

	// Does this operation hide the given op?
	public boolean hides(OperationDecl op) {
    	return false;
    }

    public boolean isMoreSpecific(OperationDecl op) {
        List myParams = parameters;
        List hisParams = op.getParameters();

        if (myParams.size() != hisParams.size())
            return false;

        Iterator myParamIter = myParams.iterator();
        Iterator hisParamIter = hisParams.iterator();
        while (myParamIter.hasNext()) {
            ParameterDecl myParam  = (ParameterDecl) myParamIter.next();
            ParameterDecl hisParam = (ParameterDecl) hisParamIter.next();
            TypeDecl myType  = myParam.getType();
            TypeDecl hisType = hisParam.getType();
            if (!hisType.assignableFrom(myType))
                return false;
		}

        return true;
	}

    public boolean applicable(List callParamTypes) {

		if (callParamTypes.size() != parameters.size())
        	return false;

		Iterator paramIter = parameters.iterator();
		Iterator callParamIter = callParamTypes.iterator();
        while (paramIter.hasNext()) {
        	ParameterDecl param = (ParameterDecl) paramIter.next();
            TypeDecl paramType = param.getType();
            if(paramType == null) {
            	return false;
            }
			TypeDecl callParamType = (TypeDecl) callParamIter.next();
            if (!paramType.assignableFrom(callParamType))
            	return false;
        }

		return true;
    }

	public PackageDecl lookupPackage(String packageName) {
    	return null; // No packages in operations.
    }

	public TypeDecl lookupType(boolean widen, String typeName, ReferenceType looker) {
    	return widen? getOwnerScope().lookupType(true, typeName, looker) : null;
    }

	public MethodDecl lookupMethod(boolean widen, String methodName, List params, ReferenceType looker) {
    	return widen? getOwnerScope().lookupMethod(true, methodName, params, looker) : null;
	}

	public VariableDecl lookupVariable(boolean widen, String name, ReferenceType looker) {

    	// Linear search of parameters
		Iterator paramIter = parameters.iterator();
        while (paramIter.hasNext()) {
        	ParameterDecl param = (ParameterDecl) paramIter.next();
			if (param.getSimpleName().equals(name))
            	return param;
        }

		return widen? getOwnerScope().lookupVariable(true, name, looker) : null;
    }

	public abstract String getSignature();
}
