package parser.grammar;

public class DottedItem implements Comparable {
    /**
     * @supplierRole production
     * @supplierCardinality 1 
     */
    protected Production production;

    protected int dotPos;

    public DottedItem(Production theProduction, int theDotPos) {
        production = theProduction;
        dotPos = theDotPos;
    }

    // Order kernel items ahead of non-kernel ones.
    public int compareTo(Object o) {
        DottedItem that = (DottedItem) o;

		if (isKernel() != that.isKernel())
			return isKernel()? -1 : 1;

        return getName().compareTo(that.getName());
    }

	public boolean isKernel() {
		return dotPos > 0 || production.getLhs().isStartSymbol();
	}

    public String getName() {
        return production.getName() + "_" + dotPos;
    }

	public Production getProduction() {
		return production;
	}

	public int getDotPos() {
		return dotPos;
	}

	public Symbol getExpectedSymbol() {
		if (dotPos < production.getSize())
            return (Symbol) production.getRhs().get(dotPos);

		return null; // Dot is at end, so don't know what to expect.
	}

	public Symbol getPriorSymbol() {
		if (dotPos > 0)
            return (Symbol) production.getRhs().get(dotPos - 1);

		return null; // Dot is at start, so don't know prior.
	}

    public String toString() {
        String result = production.getLhs().toString() + " ::=";

        int i;
        for (i = 0; i < production.getSize(); i++) {
            result += " ";

            if (i == dotPos)
                result += ". ";

            Symbol s = (Symbol) production.getRhs().get(i);
	        result += s.toString();
        }

        if (i == dotPos)
            result += " .";

        return result;
    }
}
