package parser.parsetree;

import parser.grammar.*;
import java.util.*;

/**
* SyntaxTreeBranch
* <p>
* There are two main ways to implement parse trees: (1) use a generic node for all
* productions, or (2) use production-specific branches.  The second requires code generation
* from the grammar. E.g. a code generator can subclass ParseTreeBranch to make specific
* parse tree branches, so for decl ::= type variable ;
*     class DeclNode extends ParseTreeBranch {
*     }
*     class DeclNode1 extends DeclNode {
*         TypeNode type;
*         VariableNode variable;
*         ParseTreeLeaf semi;
*     }
*/
public interface SyntaxTreeBranch extends SyntaxTree {

    /** Getter for SymbolType. */
    public Nonterminal getNonterminalType();

    /** Getter for productionType.  May be null if ambiguous or list. */
    public Production getProductionType();

    /**
     * Getter for the kid nodes.
     *
     * @return array containing SyntaxTreeNodes.
     */
    public SyntaxTree[] getKids();

    /** Tell this node it is ambiguous. */
    public void ambiguity(SyntaxTreeBranch alternate);

    public boolean isAmbiguous();

    public boolean isList();
}
