package parser.parsetree.tokenFactory;

import parser.grammar.*;
import parser.parsetree.*;

import java.util.*;

/**
* A TokenFactory that produces Tokens. These Tokens are chained together,
* so that comments and whitespace can be retained in the token sequence while not
* directly participating in parse trees.
*/
public class SimpleTokenFactory implements TokenFactory {

    /** Last token returned, for chaining. */
	protected Token prevToken;

    /**
     * Create a token with the given type, value, line and column.
     *
     * @param type the grammar terminal type of this token.
     * @param value the text string for the token.
     * @param line the line on which the token begins.
     * @param column the column on which the token begins.
     * @return a new Token or ValueToken, as necessary.
     */
    public Token makeToken(Terminal type, String value, int line, int column) {
        Token token = null;

        if (value.equals(type.toString()))
            token = new Token(type, line, column);
        else
            token = new ValueToken(type, value, line, column);

        link(token);

        return token;
    }

    /** Helper function to chain given token to previous, & make it the previous. */
	protected void link(Token next) {
	    if (prevToken != null)
	        prevToken.setNext(next);

        prevToken = next;
	}
}
