package parser.parsetree.visitor;
import java.util.*;
import parser.parsetree.*;

public abstract class SyntaxTreeVisitor {

    protected Set inProcess;

    protected SyntaxTreeVisitor() {
        inProcess = new HashSet();
    }

    public void visitNode(SyntaxTree node) {
    }

    public void visitBranch(SyntaxTreeBranch branch) {
        if (inProcess.contains(branch))
            return; // Cycle

        inProcess.add(branch);

        visitNode(branch);

        SyntaxTree[] kids = branch.getKids();
        for (int i = 0; i < kids.length; i++)
            if (kids[i] != null)
                kids[i].accept(this);

        inProcess.remove(branch);
    }

    public void visitLeaf(SyntaxTreeLeaf leaf) {
        visitNode(leaf);
    }
}
