package parser.pda;

import java.util.*;
import parser.pda.tokenStream.Looker;
import parser.parsetree.Token;
import parser.pda.stateMachine.State;
import parser.parsetree.treeFactory.SyntaxTreeFactory;
import parser.parsetree.SyntaxTree;
import parser.grammar.Production;
import parser.parsetree.*;
import util.*;
import parser.pda.stateMachine.StateMachine;

public class NondeterministicParser extends GraphDrivenParser {

    /**
     * @link aggregation
     *      @associates parser.pda.GSSNode
     * @supplierCardinality 1..*
     * @supplierRole tops
     */
    protected Map tops; // State ==> GSSNode

    /**
     * @link aggregation
     *      @associates parser.pda.GSSNode
     * @supplierCardinality 1..*
     * @supplierRole shifts
     */
    protected Map shifts; // State ==> GSSNode

    /**
     * @link aggregation
     *      @associates parser.pda.GSSNode
     * @supplierCardinality 1..*
     * @supplierRole gotos
     */
    protected Map gotos; // Pair<PdaNode, State> ==> GSSNode

    //protected Set todoGotos; // StateStackNodes to be actioned.
//public static int counter = 0;

    public NondeterministicParser(StateMachine machine, SyntaxTreeFactory theTreeBuilder) {
        super(machine, theTreeBuilder);

        tops = new HashMap();
        shifts = new HashMap();
        gotos = new HashMap();
        //todoGotos = new HashSet();
    }

    protected void start() {
        tops.clear();
        gotos.clear();
        //todoGotos.clear();

        State startState = stateMachine.getStartState();
        makeStartNode(startState);
    }

    protected void makeStartNode(State state) {
        //return new GSSNode(this, state);
        ShiftNode top = new ShiftNode(this, state);
        tops.put(state, top);
    }

	protected boolean step(Looker looker) {
        shifts.clear();
        gotos.clear();
        //todoGotos.clear();

        Iterator nodeIter = tops.values().iterator();
        while (nodeIter.hasNext()) {
            GSSNode node = (GSSNode) nodeIter.next();
            node.act(looker);
        }

        /*
        while (!todoGotos.isEmpty()) {
            GSSNode goTo = (GSSNode) todoGotos.iterator().next();

            if (counter++ == 10000) {
                counter = 0;
                System.out.println("todoGotos " + todoGotos.size());
            }
    
            goTo.act(looker);
            todoGotos.remove(goTo);
        }
        */

        if (shifts.isEmpty())
            return false;

        tops.clear();
        tops.putAll(shifts);
        return true;
    }

    protected boolean accept() {
        return tops.get(stateMachine.getStopState()) != null;
    }

    public void shift(GSSNode prev, Token input, State dst) {
        GSSNode shift = (GSSNode) shifts.get(dst);

        if (shift != null)
            shift.addPrevious(prev, input);
        else {
            shift = prev.makeShiftNode(dst, input);
            shifts.put(dst, shift);
        }
    }

    public void goTo(GSSNode prev, State dst, Production prod, List kids, Looker input) {
        assert prev != null;
        assert dst != null;
        assert prod != null;
        assert kids != null;

        GSSNode goTo = (GSSNode) gotos.get(new Pair(prev, dst));

        if (goTo == null) {
            goTo = prev.checkCycle(dst);
			if (goTo != null)
				System.out.println("Detected infinitely ambiguous goto CYCLE " + goTo);
		}

        SyntaxTreeBranch tree = treeBuilder.makeBranch(prod, kids);

        if (goTo != null) {
            SyntaxTree oldTree = goTo.getTree(prev);
            treeBuilder.makeAmbiguity(oldTree.asBranch(), tree);
        }
        else {
            goTo = prev.makeGotoNode(dst, tree);
            gotos.put(new Pair(prev, dst), goTo);
            goTo.act(input);
            //todoGotos.add(goTo); // Breadth first.
        }
    }

    protected SyntaxTreeBranch getResult() {
        ShiftNode top = (ShiftNode) tops.get(stateMachine.getStopState());

        return top == null? null : top.getResult();
    }

    protected Set getExpected() {
        Set expected = new HashSet();

        Set nodes = new HashSet();
        nodes.addAll(tops.values());
        nodes.addAll(gotos.values());

        Iterator nodeIter = nodes.iterator();
        while (nodeIter.hasNext()) {
            StateStackNode node = (StateStackNode) nodeIter.next();
            expected.addAll(node.getExpected());
        }

        return expected;
    }

    protected String getContext() {
        String result = "";

        Set nodes = new HashSet();
        nodes.addAll(tops.values());
        nodes.addAll(gotos.values());

        Iterator nodeIter = nodes.iterator();
        while (nodeIter.hasNext()) {
            StateStackNode node = (StateStackNode) nodeIter.next();

            result += node.getContext();
        }

        return result;
    }
}
