package parser.pda.stateMachine;

import parser.parsetree.*;
import parser.pda.tokenStream.*;
import parser.grammar.*;

import java.util.*;
import parser.pda.*;
import parser.pda.PdaActor;

public class NondeterministicKLookahead extends KLookahead {

    protected Set actions; // ShiftOrReduces

    /**
     * @link aggregation
     * @associates parser.pda.stateMachine.NondeterministicKLookahead
     * @supplierRole next
     * @supplierCardinality 0..*
     */
    protected Map next; // TerminalSymbol ==> NondeterministicKLookahead

    public NondeterministicKLookahead() {
        actions = new HashSet();
        next = new HashMap();
    }

    public void addAction(ShiftOrReduce action) {
        assert action != null;
        actions.add(action);
    }

    public void act(Looker input, PdaActor pda) {
        if (next.isEmpty()) {

            Iterator actionIter = actions.iterator();
            while (actionIter.hasNext()) {
                ShiftOrReduce action = (ShiftOrReduce) actionIter.next();

                act(action, input, pda);
            }

            return;
        }

        Terminal terminal = input.look();
		NondeterministicKLookahead kid = (NondeterministicKLookahead) next.get(terminal);
		if (kid != null)
			kid.act(input, pda);
    }

    public KLookahead grow(Terminal terminal) {
        KLookahead result = (NondeterministicKLookahead) next.get(terminal);

        if (result == null) {
            result = new NondeterministicKLookahead();
            next.put(terminal, result);
        }

        return result;
    }

    protected Map getNext() {
        return next;
    }

    public String toString() {
        String result = "look(";

        Iterator actionIter = actions.iterator();
        while (actionIter.hasNext()) {
            ShiftOrReduce action = (ShiftOrReduce) actionIter.next();

            if (action.isShift())
                result += " shift " + action.toString();
            else
                result += " reduce " + action.toString();
        }

        Iterator nextIter = getNext().entrySet().iterator();
        while (nextIter.hasNext()) {
            Map.Entry entry = (Map.Entry) nextIter.next();
            Terminal terminal = (Terminal) entry.getKey();
            KLookahead look = (KLookahead) entry.getValue();

            result += " on " + terminal.toString() + " ";
            result += look.toString();
        }

        result += ")";
        return result;
    }
}
