package parser.pda.stateMachine.builder;

import parser.parsetree.*;
import parser.grammar.*;
import parser.pda.stateMachine.*;
import parser.pda.stateMachine.Lookahead;

import java.util.*;

public class KLookaheadBuilder extends LookaheadBuilder {

    protected Stack lookaheads; // KLookaheads.

    protected boolean deterministic;

    public KLookaheadBuilder(boolean isDeterministic) {
        lookaheads = new Stack();
        deterministic = isDeterministic;
        reset();
    }

    public void reset() {
        lookaheads.clear();
        KLookahead root;
        if (deterministic)
            root = new DeterministicKLookahead();
        else
            root = new NondeterministicKLookahead();

        lookaheads.push(root);
    }

    public void begin(Terminal on) {
        KLookahead top = (KLookahead) lookaheads.peek();
        lookaheads.push(top.grow(on));
    }

    public void addAction(ShiftOrReduce action) {
        KLookahead top = (KLookahead) lookaheads.peek();

        if (lookaheads.size() == 1 && action.isShift())
            top = top.grow(action.asShift()); // Don't push it.

        top.addAction(action);
    }

    public void end() {
        lookaheads.pop();
    }

    public Lookahead getResult() {
        assert lookaheads.size() == 1;
        return (KLookahead) lookaheads.peek();
    }
}
