package parser.pda.stateMachine.builder;

import parser.parsetree.*;
import parser.grammar.*;
import parser.pda.stateMachine.*;

import java.util.*;

public class StateMachineBuilder {
    /**
     * @supplierRole lookBuilder
     * @supplierCardinality 1 
     */
    protected LookaheadBuilder lookBuilder;

    protected Map states; // String ==> RunState
    protected RunState currentState;

    protected boolean madeLookahead;

    protected StateMachine result;

    public StateMachineBuilder(LookaheadBuilder theLookBuilder) {
        lookBuilder = theLookBuilder;
        states = new HashMap();
        currentState = null;
        madeLookahead = false;
    }

    public void state(String name, Symbol symbol) {
        if (!states.containsKey(name))
            states.put(name, new RunState(symbol));
    }

    public void beginState(String name) {
        assert currentState == null;
        currentState = (RunState) states.get(name);
        madeLookahead = false;
    }

	public void addItem(DottedItem item) {
		currentState.addItem(item);
	}

    public void transition(String stateName) {
        currentState.addTransition((RunState) states.get(stateName));
    }

    public void reduction(Production prod) {
        currentState.addReduction(prod);
    }

    public void beginOn(Terminal on) {
        lookBuilder.begin(on);
    }

    public void addAction(ShiftOrReduce action) {
        lookBuilder.addAction(action);
        madeLookahead = true;
    }

    public void endOn() {
        lookBuilder.end();
    }

    public void endState() {
        assert currentState != null;

        if (madeLookahead) {
            Lookahead look = lookBuilder.getResult();
            currentState.setLookahead(look);
    
            lookBuilder.reset();
        }

        currentState = null;
        madeLookahead = false;
    }

    public void makeMachine(String startName, String stopName) {
        State startState = (State) states.get(startName);
        State stopState = (State) states.get(stopName);

        result = new StateMachine(startState, stopState);
    }

    /*
    public State getState(String name) {
        return (State) states.get(name);
    }
    */

    public StateMachine getResult() {
        return result;
    }
}
